/*
 * Copyright (c) 2007-2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Wed Jan 24 2007
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/stat.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include "opencv2/core/core_c.h"
#endif


/* --- local data ---------------------------------------------------- */

static viam_acqdev_obj	viam_hwfind(viam_handle_t vh, const char *name);
static viam_acqdev_obj	viam_hwload(viam_handle_t vh, const char *name);
static const char *	viam_hwdpath(void);


static struct viam_driver_obj {
  void *dlhandle;
  viam_acqdev_obj hwdev;
} *hwdev_list = NULL;			/**< loaded drivers array */

static unsigned int nhwdevs = 0;	/**< number of loaded drivers */


/* --- viam_hwdev_name ----------------------------------------------- */

const char *
viam_hwdev_name(viam_acqdev_obj hwdev)
{
  return hwdev->name;
}


/* --- viam_hardware_load -------------------------------------------- */

/** \brief Load a specific hardware driver
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Driver name.
 *
 * \return error code.
 */

int
viam_hardware_load(viam_handle_t vh, const char *driver)
{
  viam_acqdev_obj h;

  h = viam_hwfind(vh, driver);
  if (!h) {
    h = viam_hwload(vh, driver);
    if (!h) return viam_seterrno(vh, Eviamlib_NO_SUCH_DEVICE);
  }

  return 0;
}


/* --- viam_hardware_unload ------------------------------------------ */

/** \brief Unload a specific hardware driver
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Driver name.
 *
 * \return error code.
 */

int
viam_hardware_unload(viam_handle_t vh, const char *driver)
{
  viam_acqdev_obj h = viam_hwfind(vh, driver);
  if (!h) return viam_seterrno(vh, Eviamlib_NO_SUCH_DEVICE);

  /* XXX unload it! */
  return 0;
}


/* --- viam_hardware_list -------------------------------------------- */

/** \brief Print available camera devices for all loaded drivers
 *
 * \param[in] vh	Library handle.
 *
 * \return error code.
 */

int
viam_hardware_list(viam_handle_t vh)
{
  struct viam_driver_obj *d;
  int i;

  for(i=0, d = hwdev_list; i < nhwdevs; i++, d++)
    if (d->hwdev->list(vh))
      warnx("cannot list hardware for %s driver",
	    viam_hwdev_name(d->hwdev));

  return 0;
}


/* --- viam_hardware_device ------------------------------------------ */

/** \brief Get device structure from is id
 *
 * \param[in] vh	Library handle.
 * \param[in] id	Camera unique id.
 * \param[out] uid	Pointer on unique id.
 *
 * \return error code.
 */

viam_acqdev_obj
viam_hardware_device(viam_handle_t vh, const char *id, char **uid)
{
  viam_acqdev_obj h;

  *uid = strchr(id, ':');
  if (!*uid) { *uid = (char *)id; return NULL; }
  (*uid)++;

  {
    char name[*uid - id];
    strncpy(name, id, *uid-id-1);
    name[*uid-id-1] = 0;

    h = viam_hwfind(vh, name);
    if (!h) {
      h = viam_hwload(vh, name);
      if (!h) {
	viam_seterrno(vh, Eviamlib_NO_SUCH_DEVICE);
	*uid = NULL;
	return NULL;
      }
    }
  }

  return h;
}


/* --- viam_hardware_attach ------------------------------------------ */

/** \brief Scan hardware and attach cameras
 *
 * \param[in] vh	Library handle.
 *
 * \return error code
 */

int
viam_hardware_attach(viam_handle_t vh)
{
  viam_acqdev_obj h;
  viam_camera_t c;
  int i;

  for (c = viam_config_cameras(vh); c; c = viam_camera_next(c)) {
    if (viam_camera_hwdata(c)) continue;

    h = viam_camera_hwdev(c);
    if (h) {
      if (h->attach(vh, c))
	warnx("camera `%s' not found on %s device",
	      viam_camera_name(c), h->name);
    } else {
      struct viam_driver_obj *d;

      warnx("camera %s without explicit device, trying all",
	    viam_camera_name(c));
      for(i=0, d=hwdev_list; i<nhwdevs; i++, d++)
	if (!d->hwdev->attach(vh, c)) {
	  viam_camera_sethwdev(c, d->hwdev);
	  break;
	} else
	  warnx("cannot attach camera %s on %s device",
		viam_camera_name(c), h->name);

      h = viam_camera_hwdev(c);
      if (!h)
	warnx("camera `%s' not found on any devices", viam_camera_name(c));
    }

    if (h)
      warnx("attached camera `%s' on %s device", viam_camera_name(c), h->name);
  }

  return 0;
}


/* --- viam_hardware_detach ------------------------------------------ */

/** \brief Cleanup hardware
 *
 * \param[in] vh	Library handle.
 *
 * \return error code
 */

void
viam_hardware_detach(viam_handle_t vh)
{
  viam_bank_t b;
  viam_camera_t c;

  for(b = viam_config_banks(vh); b; b = viam_bank_next(b))
    viam_bank_unconfigure(vh, b);

  for (c = viam_config_cameras(vh); c; c = viam_camera_next(c))
    if (viam_camera_hwdev(c)) {
      viam_camera_hwdev(c)->detach(vh, c);
      viam_camera_setsetup(c, 0);
    }
}


/* --- viam_bank_configure ------------------------------------------- */

/** \brief Initialize cameras attached to bank
 *
 * \param[in] vh	Library handle.
 * \param[in] bank	Bank handle.
 *
 * \return error code
 */

int
viam_bank_configure(viam_handle_t vh, viam_bank_t bank)
{
  viam_camera_t camera;
  viam_image_ref im;
  int i, j, s;

  if (viam_bank_isconfigured(bank))
    return viam_seterrno(vh, Eviamlib_CONFIGURED);

  /* iterate over all cameras */
  for(i = 0; i < viam_bank_ncameras(bank); i++) {
    camera = viam_bank_camera(vh, bank, i);
    if (!viam_camera_hwdev(camera)) continue;
    if (viam_camera_setup(camera)) continue;

    s = viam_camera_hwdev(camera)->configure(vh, camera);
    if (s) {
      for(j = 0; j < i; j++) {
	camera = viam_bank_camera(vh, bank, j);
	viam_camera_hwdev(camera)->unconfigure(vh, camera);
	viam_camera_setsetup(camera, 0);
      }
      return s;
    }
    viam_camera_setsetup(camera, 1);
  }

  /* configure filters */
  for(i = 0; i < viam_bank_nimages(bank); i++) {
    im = viam_bank_image(vh, bank, i);

    s = viam_filter_configure(vh, im, viam_image_flist(im));
    if (s) {
      warnx("image `%s' has misconfigured filters", viam_image_name(im));
      return viam_seterrno(vh, s);
    }
  }

  viam_bank_setconfigured(bank, 1);
  return 0;
}


/* --- viam_bank_unconfigure ----------------------------------------- */

/** \brief Unconfigure bank
 *
 * \param[in] vh	Library handle.
 * \param[in] bank	Bank handle.
 *
 * \return error code
 */

int
viam_bank_unconfigure(viam_handle_t vh, viam_bank_t bank)
{
  viam_camera_t camera;
  int i;

  if (!viam_bank_isconfigured(bank)) return 0;

  /* iterate over all cameras */
  for(i = 0; i < viam_bank_ncameras(bank); i++) {
    camera = viam_bank_camera(vh, bank, i);
    if (!viam_camera_setup(camera)) continue;

    viam_camera_hwdev(camera)->unconfigure(vh, camera);
    viam_camera_setsetup(camera, 0);
  }

  viam_bank_setconfigured(bank, 0);
  return 0;
}


/* --- viam_datatransmit --------------------------------------------- */

/** \brief Start or stop isochronous data transmission
 *
 * \param[in] vh	Library handle
 * \param[in] bank	Bank handle
 * \param[in] on	Start transmission if true, stop if false.
 *
 * \return 0 on success
 */

int
viam_datatransmit(viam_handle_t vh, viam_bank_t bank, viam_onoff_t on)
{
  viam_camera_t c;
  int i, j, s;

  if (!viam_bank_isconfigured(bank))
    return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  if (on) {
    /* initialize hardware filters */
    for(i=0; i<viam_bank_nimages(bank); i++) {
      s = viam_filter_inithw(vh, viam_bank_image(vh, bank, i));
      if (s) return s;
    }
  }

  /* turn on/off data transmission */
  for(i=0; i<viam_bank_ncameras(bank); i++) {
    c = viam_bank_camera(vh, bank, i);

    s = viam_camera_hwdev(c)->datatransmit(vh, c, on);
    if (s && on) {
      for(j = 0; j < i; j++) {
        c = viam_bank_camera(vh, bank, j);
        viam_camera_hwdev(c)->datatransmit(vh, c, VIAM_OFF);
      }
      return s;
    }
  }

  return 0;
}


/* --- viam_oneshot -------------------------------------------------- */

/** \brief Shoot each camera of a bank
 *
 * \param[in] vh	Library handle.
 * \param[in] bank	Bank reference.
 * \param[in] image	User-supplied array images, with allocated buffer.
 * \param[in] timestamps User-supplied array of image acquisition
 *			timestamps.
 * \param[in] nimages	Number of images.
 *
 * \return 0 on success
 */

int
viam_oneshot(viam_handle_t vh, viam_bank_t bank, IplImage *image[],
	     struct timeval *timestamps[], int nimages)
{
  viam_camera_t c;
  viam_image_ref im;
  viam_geometry_t geom;
  size_t bytes, stride;
  char *buffer;
  int i, s;

  if (!viam_bank_isconfigured(bank))
    return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
  if (nimages != viam_bank_nimages(bank))
    return viam_seterrno(vh, Eviamlib_BAD_IMAGE_COUNT);

  /* capture all cameras */
  for(i=0; i<viam_bank_ncameras(bank); i++) {
    c = viam_bank_camera(vh, bank, i);
    s = viam_camera_hwdev(c)->capture(vh, c);
    if (s) return s;
  }

  /* get images */
  for(i=0; i<viam_bank_nimages(bank); i++) {
    im = viam_bank_image(vh, bank, i);
    c = viam_image_camera(im);

    buffer = viam_camera_hwdev(c)->capturedata(vh, c, &bytes, &stride);
    if (!buffer) goto done;

    cvSetData(viam_camera_raw(c), buffer, stride);

    s = viam_image_geometry(vh, im, &geom);
    if (s) goto done;
    if (image[i]->width != geom.width || image[i]->height != geom.height ||
	image[i]->depth != geom.depth || image[i]->nChannels != geom.channels) {
      s = viam_seterrno(vh, Eviamlib_IMAGE_GEOMETRY);
      goto done;
    }

    s = viam_filter_image(vh, im, image[i]);
    if (s) goto done;

    if (timestamps)
      *timestamps[i] = viam_camera_hwdev(c)->capturetimestamp(vh, c);
  }

done:
  for(i=0; i<viam_bank_ncameras(bank); i++) {
    c = viam_bank_camera(vh, bank, i);
    viam_camera_hwdev(c)->capturedone(vh, c);
  }

  return 0;
}


/* --- viam_queryhwmodes --------------------------------------------- */

/** \brief Query available modes of a camera
 *
 * \param[in] vh		Library handle.
 * \param[in] camera		Camera handle.
 * \param[out] supported	Array of supported geometries.
 * \param[in] n			Maximum number of geometries.
 *
 * \return 0 on success.
 */

int
viam_queryhwmodes(viam_handle_t vh, viam_camera_t camera, int offset,
		  viam_hwmode_t *supported, size_t n)
{
  return
    viam_camera_hwdev(camera)->queryhwmodes(vh, camera, offset, supported, n);
}


/* --- viam_queryfeature --------------------------------------------- */

/** \brief Query feature capabilities of a camera.
 *
 * \param[in] vh		Library handle.
 * \param[in] camera		Camera handle.
 * \param[in] filter		Feature type.
 * \param[out] methods		Or'ed combination of supported methods.
 * \param[out] aumodes		Or'ed combination of supported auto modes.
 *
 * \return 0 on success.
 */

int
viam_queryfeature(viam_handle_t vh,
		  viam_camera_t camera, viam_filtertype_t filter,
		  unsigned int *methods, unsigned int *aumodes)
{
  unsigned int hwm, hwa, swm, swa;
  int s;

  s = viam_camera_hwdev(camera)->queryhwfeature(vh, camera, filter, &hwm, &hwa);
  if (s) return s;

  s = viam_filter_queryswfeature(vh, filter, &swm, &swa);
  if (s) return s;

  if (methods) *methods = hwm | swm;
  if (aumodes) *aumodes = hwa | swa;
  return 0;
}


/* --- viam_sethwfeaturemode ----------------------------------------- */

/** \brief Set harware feature capabilities of a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] method	Filter application method.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

int
viam_sethwfeaturemode(viam_handle_t vh, viam_camera_t camera,
		      viam_filtertype_t filter, viam_filtermethod_t method,
		      viam_filterautomode_t aumode)
{
  return viam_camera_hwdev(camera)->
    sethwfeaturemode(vh, camera, filter, method, aumode);
}


/* --- viam_sethwfeaturevalue ---------------------------------------- */

/** \brief Set hardware feature value for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] args	Feature value (NULL for AUTO or ONE_PUSH).
 *
 * \return 0 on success.
 */

int
viam_sethwfeaturevalue(viam_handle_t vh, viam_camera_t camera,
		       viam_filtertype_t filter, const viam_filterargs_obj args)
{
  return viam_camera_hwdev(camera)->
    sethwfeaturevalue(vh, camera, filter, args);
}


/* -- viam_hwmode_size ----------------------------------------------- */

/** \brief Return hardware mode size.
 *
 * The function returns parameters compatible with cvCreateImage()
 * function.
 *
 * \param[in] vh	Library handle.
 * \param[in] hwmode	Hardware mode.
 * \param[out] width	Width in pixels.
 * \param[out] height	Height in pixels.
 * \param[out] depth	Bit depth of pixels.
 * \param[out] channels	Number of channels per elements.
 *
 * \return 0 if hwmode is known and valid, with corresponding non NULL
 * out parameters instanciated.
 */

int
viam_hwmode_size(viam_handle_t vh, const viam_hwmode_t *hwmode,
		 int *width, int *height, int *depth, int *channels)
{
#define setmaybe(x, v)	do if (x) *(x) = (v); while(0)
  switch(hwmode->size) {
    case VIAM_HWSZ_INVALID:
      return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

    case VIAM_HWSZ_160x120:
      setmaybe(width, 160);	setmaybe(height, 120);	break;
    case VIAM_HWSZ_320x240:
      setmaybe(width, 320);	setmaybe(height, 240);	break;
    case VIAM_HWSZ_512x384:
      setmaybe(width, 512);	setmaybe(height, 384);	break;
    case VIAM_HWSZ_640x480:
      setmaybe(width, 640);	setmaybe(height, 480);	break;
    case VIAM_HWSZ_800x600:
      setmaybe(width, 800);	setmaybe(height, 600);	break;
    case VIAM_HWSZ_1024x768:
      setmaybe(width, 1024);	setmaybe(height, 768);	break;
    case VIAM_HWSZ_1280x960:
      setmaybe(width, 1280);	setmaybe(height, 960);	break;
    case VIAM_HWSZ_1600x1200:
      setmaybe(width, 1600);	setmaybe(height, 1200);	break;
  }

  switch(hwmode->format) {
    case VIAM_HWFMT_INVALID:
      return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

    case VIAM_HWFMT_MONO16:
      setmaybe(depth, IPL_DEPTH_16U);	setmaybe(channels, 1);	break;
    case VIAM_HWFMT_MONO8:
    case VIAM_HWFMT_YUV411:
    case VIAM_HWFMT_YUV422_UYVY:
    case VIAM_HWFMT_YUV422_YUYV:
    case VIAM_HWFMT_YUV422_YYUV:
    case VIAM_HWFMT_YVU422_VYUY:
    case VIAM_HWFMT_YVU422_YVYU:
      setmaybe(depth, IPL_DEPTH_8U);	setmaybe(channels, 1);	break;
    case VIAM_HWFMT_YUV444:
    case VIAM_HWFMT_RGB888:
      setmaybe(depth, IPL_DEPTH_8U);	setmaybe(channels, 3);	break;
    case VIAM_HWFMT_F32:
      setmaybe(depth, IPL_DEPTH_32F);	setmaybe(channels, 1);	break;
  }

#undef setmaybe
  return 0;
}


/* --- viam_hwfind --------------------------------------------------- */

/** \brief Return a hardware driver.
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Driver name.
 *
 * \return the driver object or NULL in case of error.
 */

static viam_acqdev_obj
viam_hwfind(viam_handle_t vh, const char *name)
{
  struct viam_driver_obj *d;
  viam_acqdev_obj hwdev = NULL;

  /* iterate over already loaded drivers */
  for(d = hwdev_list; d - hwdev_list < nhwdevs; d++) {
    if (strcmp(viam_hwdev_name(d->hwdev), name)) continue;

    hwdev = d->hwdev;
    break;
  }

  return hwdev;
}


/* --- viam_hwload --------------------------------------------------- */

/** \brief Load a hardware driver.
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Driver name.
 *
 * \return the driver object or NULL in case of error.
 */

static viam_acqdev_obj
viam_hwload(viam_handle_t vh, const char *name)
{
  const char *dpath = viam_hwdpath();
  struct viam_driver_obj drv = { NULL, NULL };

  struct stat sb;
  char path[strlen(dpath)+1];
  char *p;

  /* look for name.so in drivers path */
  strcpy(path, dpath);
  for(p = strtok(path, ":"); p; p = strtok(NULL, ":")) {
    char file[strlen(p)+1/*/*/+strlen(name)+3/*.so*/+1/*\0*/];

    snprintf(file, sizeof(file), "%s/%s.so", p, name);
    if (stat(file, &sb)) continue;

    drv.dlhandle = dlopen(file, RTLD_LAZY);
    if (!drv.dlhandle) {
      warnx("cannot load %s: %s", file, dlerror());
      continue;
    }

    drv.hwdev = dlsym(drv.dlhandle, viam_acqdev_sym);
    if (!drv.hwdev) {
      dlclose(drv.dlhandle);
      warnx("no viam driver in %s", file);
      continue;
    }

    break;
  }
  if (!drv.hwdev) { warnx("no driver for %s", name); return NULL; }

  /* initalize driver */
  if (drv.hwdev->init)
    if (drv.hwdev->init(vh)) {
      dlclose(drv.dlhandle);
      warnx("cannot initialize %s driver: disabling", name);
      return NULL;
    }

  /* register in global drivers array */
  {
    struct viam_driver_obj *n;

    n = realloc(hwdev_list, (nhwdevs+1) * sizeof(*hwdev_list));
    if (!n) { warnx("cannot register %s", name); return NULL; }

    hwdev_list = n;
    hwdev_list[nhwdevs++] = drv;
  }

  warnx("loaded %s driver", name);
  return drv.hwdev;
}


/* --- viam_hwdpath -------------------------------------------------- */

/** \brief Return drivers path.
 *
 * \return the drivers path.
 */

static const char *
viam_hwdpath()
{
#ifndef VIAM_DEFAULT_DRIVER_PATH
# error "no default driver path defined in VIAM_DEFAULT_DRIVER_PATH"
#endif
  const char *dpath = getenv("VIAM_DRIVER_PATH");

  return dpath?dpath:VIAM_DEFAULT_DRIVER_PATH;
}
